import asyncio
import logging
import shutil
from collections import OrderedDict
from typing import Any, Coroutine, Dict, List, Text

from sanic import Blueprint, response
from sanic.request import Request

import rasax.community.version
from rasax.community import config, utils
from rasax.community.api.decorators import inject_rasa_x_user, platform_scoped
from rasax.community.services.feature_service import FeatureService
from rasax.community.services.projects_service import ProjectsService
from rasax.community.services.settings_service import SettingsService
from rasax.community.services.stack_service import StackService
from rasax.community.services.user_service import (
    UserException, UserService,
    MismatchedPasswordsException)
from rasax.community.utils import error, dump_yaml

logger = logging.getLogger(__name__)


def collect_status_calls(service_dict: Dict[Text, StackService]
                         ) -> List[Coroutine]:
    """Returns a list of status() Coroutines.

    Creates an OrderedDict of statuses from all stack services."""

    calls_dict = OrderedDict(
        [(name, service.status()) for name, service in service_dict.items()]
    )
    return list(calls_dict.values())


def _stack_services(request: Request) -> Dict[Text, StackService]:
    settings_service = SettingsService(request["db_session"])
    return settings_service.stack_services()


async def collect_stack_results(stack_services: Dict[Text, StackService],
                                statuses: OrderedDict
                                ) -> Dict[Text, Any]:
    """Creates status result dictionary for stack services."""

    stack_result = dict()

    for name, _status in zip(stack_services.keys(), statuses):
        if isinstance(_status, dict) and "version" in _status:
            _result = _status.copy()
            _result["status"] = 200
        else:
            _result = {"status": 500, "message": _status}
        stack_result[name] = _result

    return stack_result


def blueprint() -> Blueprint:
    project_endpoints = Blueprint('project_endpoints')

    # noinspection PyUnusedLocal
    @project_endpoints.route("/health", methods=['GET', 'HEAD'])
    # @platform_scoped('project.health.get', user_service=user_service)
    async def status(request):
        stack_services = _stack_services(request)
        stack_status_calls = collect_status_calls(stack_services)
        statuses = await asyncio.gather(*stack_status_calls,
                                        return_exceptions=True)
        stack_result = await collect_stack_results(stack_services, statuses)

        return response.json(stack_result)

    # noinspection PyUnusedLocal
    @project_endpoints.route("/version", methods=['GET', 'HEAD'])
    # @platform_scoped('project.version.get', user_service=user_service)
    async def version(request):
        stack_services = _stack_services(request)
        stack_status_calls = collect_status_calls(stack_services)

        statuses = await asyncio.gather(*stack_status_calls,
                                        return_exceptions=True)
        stack_versions = dict()
        for name, _status in zip(stack_services.keys(), statuses):
            if isinstance(_status, dict) and "version" in _status:
                _result = _status.get("version")
            else:
                _result = {"status": 500, "message": _status}
            stack_versions[name] = _result

        result = {
            "stack": stack_versions,
            "platform": rasax.community.__version__
        }
        return response.json(result)

    # noinspection PyUnusedLocal
    @project_endpoints.route("/user", methods=['GET', 'HEAD'])
    @platform_scoped('user.get')
    @inject_rasa_x_user()
    async def profile(request, user=None):
        user_service = UserService(request["db_session"])
        return response.json(user_service.fetch_user(user["username"],
                                                     return_api_token=True))

    @project_endpoints.route("/user", methods=['PATCH'])
    @platform_scoped('user.update')
    @inject_rasa_x_user()
    async def update_username(request, user=None):
        rjs = request.json
        if not utils.check_schema("username", rjs):
            return error(401, "WrongSchema",
                         "Failed to validate schema for username payload")

        try:
            user_service = UserService(request["db_session"])
            user_profile = user_service.update_saml_username(user["name_id"],
                                                             rjs["username"])

        except UserException as e:
            return error(
                404, "UserException",
                "Could not assign username {} to name_id {}"
                "".format(rjs["username"], user["name_id"]),
                details=e
            )

        return response.json(user_profile)

    @project_endpoints.route("/users", methods=['GET', 'HEAD'])
    @platform_scoped('users.list')
    async def list_users(request):
        user_service = UserService(request["db_session"])
        users = user_service.fetch_all_users(config.team_name)
        if not users:
            return error(404, "NoUsersFound", "No users found")

        profiles = [user_service.fetch_user(u["username"]) for u in users]

        return response.json(profiles,
                             headers={"X-Total-Count": len(profiles)})

    @project_endpoints.route("/users", methods=['POST'])
    @platform_scoped('users.create')
    async def create_user(request):
        rjs = request.json
        if not utils.check_schema("user", rjs):
            return error(401, "WrongSchema",
                         "Failed to validate schema for user payload")

        user_service = UserService(request["db_session"])
        team = rjs.get("team") or config.team_name
        user_service.create_user(username=rjs["username"],
                                 raw_password=rjs["password"],
                                 team=team,
                                 roles=rjs["roles"])

        return response.json(rjs, 201)

    @project_endpoints.route("/users/<username>", methods=['DELETE'])
    @platform_scoped('users.delete')
    async def delete_user(request, username):
        user_service = UserService(request["db_session"])

        try:
            deleted = user_service.delete_user(username)
            return response.json(deleted)
        except UserException as e:
            return response.json({"error": "{}".format(e)})

    @project_endpoints.route("/user/password", methods=['POST'])
    @platform_scoped('user.password.update')
    async def change_password(request):
        rjs = request.json
        if not utils.check_schema("change_password", rjs):
            return error(401, "WrongSchema",
                         "Failed to validate schema for password payload")

        user_service = UserService(request["db_session"])

        try:
            user = user_service.change_password(rjs)
            if user is None:
                return error(404, "UserNotFound", "user not found")
            return response.json(user)
        except MismatchedPasswordsException:
            return error(403, "WrongPassword", "wrong password")

    # noinspection PyUnusedLocal
    @project_endpoints.route("/projects/<project_id>", methods=['POST'])
    @platform_scoped('projects.create')
    @inject_rasa_x_user()
    async def create_project(request, project_id, user=None):
        # TODO: check if project exists for saving it to avoid
        # db errors
        settings_service = SettingsService(request["db_session"])
        project = settings_service.init_project(user["team"], project_id)

        user_service = UserService(request["db_session"])
        user_service.create_project(user, project_id)

        return response.json(project)

    # no authentication because features may be needed
    # before a user is authenticated¬
    @project_endpoints.route("/features", methods=['GET', 'HEAD'])
    async def features(request):
        feature_service = FeatureService(request["db_session"])
        return response.json(feature_service.features())

    @project_endpoints.route("/features", methods=['POST'])
    @platform_scoped('features.create', allow_api_token=True)
    async def set_feature(request):
        rjs = request.json
        if not utils.check_schema("feature", rjs):
            return error(401, "WrongSchema",
                         "Failed to validate schema for feature payload")

        feature_service = FeatureService(request["db_session"])
        feature_service.set_feature(rjs)
        return response.json(rjs)

    # noinspection PyUnusedLocal
    @project_endpoints.route("/logs")
    @platform_scoped('logs.list', allow_api_token=True)
    async def logs(_):
        try:
            shutil.make_archive("/tmp/logs", "zip", "/logs")
            return await response.file('/tmp/logs.zip')
        except Exception as e:
            logger.debug(e)
            return error(500, "error collecting logs", e)

    # noinspection PyUnusedLocal
    @project_endpoints.route("/environments", methods=['GET', 'HEAD'])
    @platform_scoped('environments.list', allow_api_token=True)
    @inject_rasa_x_user()
    async def get_environment_config(request, user=None):
        settings_service = SettingsService(request["db_session"])
        environments = settings_service.get_environments_config(
            config.project_name)

        if not environments:
            return error(400, "EnvironmentSettingsNotFound",
                         "could not find environment settings")

        return response.json(
            {"environments": dump_yaml(environments.get("environments"))}
        )

    @project_endpoints.route("/chatToken", methods=['GET'])
    @platform_scoped('chatToken.get', allow_rasax_token=True)
    async def get_chat_token(request):
        projects_service = ProjectsService(request["db_session"])
        chat_token = projects_service.get_token()
        return response.json({"bot_name": chat_token.bot_name,
                              "description": chat_token.description,
                              "chat_token": chat_token.token})

    @project_endpoints.route("/chatToken", methods=['PUT'])
    @platform_scoped('chatToken.update', allow_api_token=True)
    async def update_chat_token(request):
        rjs = request.json
        if not utils.check_schema("update_token", rjs):
            return error(401, "WrongSchema",
                         "Failed to validate schema for update_token payload")
        bot_name = rjs.get("bot_name", "")
        description = rjs.get("description", "")
        projects_service = ProjectsService(request["db_session"])
        projects_service.update_token(bot_name=bot_name,
                                      description=description)
        chat_token = projects_service.get_token()
        return response.json({"bot_name": bot_name, "description": description,
                              "chat_token": chat_token.token})

    return project_endpoints
