import json
import logging
from typing import Dict, Text, Any, Optional

from ruamel.yaml.compat import ordereddict
from sqlalchemy import and_
from sqlalchemy.orm import Session

from rasa.utils.io import read_yaml_file
from rasax.community import config
from rasax.community.database import Environment, Project, LocalPassword
from rasax.community.services.data_service import DataService
from rasax.community.services.projects_service import ProjectsService
from rasax.community.services.stack_service import (
    StackService,
    AsyncEndpointConfig)
from rasax.community.services.story_service import StoryService

logger = logging.getLogger(__name__)


def default_environments_config():
    stack_config = ordereddict(
        [("production", ordereddict(
            [("url", "http://stack:5005"),
             ("token", config.stack_token),
             ("db", "rasa")])),
         ("development", ordereddict(
             [("url", "http://stack-development:5005"),
              ("token", config.stack_token),
              ("db", "rasa_development")])),
         ("worker", ordereddict(
             [("url", "http://stack-worker:5005"),
              ("token", config.stack_worker_token),
              ("db", "rasa")]))])

    return {"environments": ordereddict([("stack", stack_config)])}


def default_environments_config_local():
    stack_config = ordereddict([
        ("production", ordereddict(
            [("url", "http://localhost:5005"),
             ("token", config.stack_token),
             ("db", "rasa")])
         ),
        ("worker", ordereddict(
            [("url", "http://localhost:5005"),
             ("token", config.stack_token),
             ("db", "rasa")])
         )
    ])

    return {"environments": ordereddict([("stack", stack_config)])}


def default_stack_config():
    policies = [{"name": "MemoizationPolicy"},
                {"name": "KerasPolicy"},
                {"name": "MappingPolicy"}]
    return ordereddict(
        [("pipeline", "supervised_embeddings"),
         ("language", "en"),
         ("policies", policies)]
    )


class SettingsService(object):

    def __init__(self, session: Session):
        self.session = session

    def init_project(self, team: Text, project_id: Text,
                     is_in_local_mode: bool = False):
        new_project = Project(project_id=project_id, team=team,
                              config=json.dumps(default_stack_config()))
        self.session.add(new_project)

        env_config = self.get_environments_config(project_id
                                                  ).get("environments", {})

        if not env_config.get("stack"):
            if is_in_local_mode:
                default_env = default_environments_config_local()
            else:
                default_env = default_environments_config()

            self.save_environments_config(project_id, default_env.get(
                "environments"))

        self.session.commit()

        return new_project.as_dict()

    def get(self, team: Text, project_id: Text) -> Optional[Dict[Text, Text]]:
        project = self._get_project(team, project_id)

        if project:
            return project.as_dict()

    def _get_project(self, team: Text, project_id: Text) -> Optional[Project]:
        return self.session.query(Project).filter(
            and_(Project.project_id == project_id,
                 Project.team == team)).first()

    def save_config(self, team: Text, project_id: Text,
                    stack_config: Dict[Text, Any]) -> Dict[Text, Any]:
        logger.debug("saving : ")
        logger.debug(stack_config)

        project = self._get_project(team, project_id)
        project.config = json.dumps(stack_config)

        self.session.commit()

        return stack_config

    def inject_environments_config_from_file(self,
                                             project_id: Text,
                                             filename: Text):
        """Inject a deployment environments configuration file at `filename`
            into the db.
        """

        try:
            _config = read_yaml_file(filename)
            self.inspect_environments_config(_config)
            self.save_environments_config(project_id, _config)
            logger.debug("Successfully injected deployment environments "
                         "configuration from file '{}'.".format(filename))
        except (ValueError, FileNotFoundError) as e:
            current_config = self.get_environments_config(project_id)
            logger.debug("Could not inject deployment environments "
                         "configuration from file '{}'. Details:\n{}\nYou may "
                         "still use Rasa Platform if you currently have a "
                         "working configuration. Your current configuration "
                         "is:\n{}".format(filename, e,
                                          json.dumps(current_config, indent=2)))

    @staticmethod
    def inspect_environments_config(env_config):
        """Inspect deployment environment config.

        Raises ValueError if production and worker services do not share the
        same db, or if db entries for services other than production and
        worker are not unique."""

        stack_config = env_config["stack"]
        production_db = stack_config["production"]["db"]
        # check production and worker dbs are the same
        for environment in ("production", "worker"):
            db = stack_config[environment]["db"]
            if db is None or db != production_db:
                raise ValueError(
                    "Database '{}' was found to be different "
                    "from the one that should be used for "
                    "production and worker services ('{}')."
                    "".format(db, production_db))

        # check other dbs are unique
        for k, v in stack_config.items():
            if k in ("production", "worker"):
                continue

            if v["db"] == production_db:
                raise ValueError(
                    "Rasa database '{}' cannot be the same one that's "
                    "used for production and worker services ('{}')."
                    "".format(db, production_db))

    def save_environments_config(self, project_id: Text,
                                 environments_config: Dict[Text, Any]
                                 ) -> Dict[Text, Dict[Text, Dict[Text, Text]]]:
        for name, env_config in environments_config.get("stack", {}).items():
            # delete possibly existing config
            self.delete_environment_config(name, project_id)

            env = Environment(name=name, project=project_id,
                              url=env_config["url"],
                              token=env_config["token"],
                              db=env_config["db"])
            self.session.add(env)
            self.session.commit()

        return self.get_environments_config(project_id)

    def delete_environment_config(self, name, project_id):
        to_delete = self.session.query(Environment).filter(
            Environment.name == name).filter(
            Environment.project == project_id).first()
        if to_delete:
            self.session.delete(to_delete)
            self.session.commit()

    def get_environments_config(self,
                                project_id: Text
                                ) -> Dict[Text, Any]:
        envs = self.session.query(Environment).filter(
            Environment.project == project_id).all()
        envs = [e.as_dict() for e in envs]

        return {"environments": {"stack": {list(e)[0]: e[list(e)[0]]
                                           for e in envs}}}

    def get_stack_service(self, environment: Text,
                          project_id: Text = config.project_name
                          ) -> Optional[StackService]:
        return self.stack_services(project_id).get(environment, None)

    def stack_services(self,
                       project_id: Text = config.project_name
                       ) -> Dict[Text, StackService]:
        """Create StackServices for all Stack servers."""
        environments_config = self.get_environments_config(
            project_id).get("environments", {})

        _stack_services = {}
        for k, v in environments_config.get("stack", {}).items():
            _config = AsyncEndpointConfig(v.get("url"), token=v.get("token"))
            _stack_services[k] = StackService(_config,
                                              DataService(self.session),
                                              StoryService(self.session),
                                              ProjectsService(self.session),
                                              self)
        return _stack_services

    def save_local_password(self, password: Text) -> None:
        """Save Rasa X password in local mode.

        Overwrite existing local password, or create a new one if none is found.
        """

        local_password = self.session.query(LocalPassword).first()
        if local_password:
            local_password.password = password
        else:
            local_password = LocalPassword(password=password)
            self.session.add(local_password)

        self.session.commit()
